/* 
 * Jason McMullan
 * 
 * Linux driver for SmartDisk products
 * $Revision: 1.17 $
 *
 * $Log: kernel.c,v $
 * Revision 1.17  2000/04/14 13:49:17  gus
 * Corel fixes
 *
 * Revision 1.16  2000/03/10 23:54:28  gus
 * Fixed stuff, and certified the driver
 *
 * Revision 1.15  2000/03/10 19:08:35  gus
 * Changed status reporting system - works better now
 *
 * Revision 1.14  2000/03/10 00:18:05  gus
 * Waiting inode mangler - works..
 *
 * Revision 1.13  2000/03/09 18:14:11  gus
 * Works... But I'm about to try to optimize for frequent re-opens
 *
 * Revision 1.12  2000/03/06 22:08:39  gus
 * Fixed the normal-user-cant-use-mtools bug
 *
 * Revision 1.11  2000/02/27 05:34:31  gus
 * Ahh.. Final fixes...
 *
 * Revision 1.10  2000/02/24 09:51:14  gus
 * ALPHA Release. This has a working fpdaemon implementation, and a
 * 	working driver implementation. Yee ha.
 *
 * Revision 1.9  2000/02/21 23:21:52  gus
 * Finally got the floppy_raw_cmd based IO working, and it works like
 * a dream! Now performing some tests to make sure I didn't screw something up.
 *
 * Revision 1.8  2000/02/18 22:55:33  jmcc
 * Latest rev - includes all 3 types of IO, but seems to kill the
 * buffer system on too many writes...
 *
 * Revision 1.7  2000/02/17 20:14:21  jmcc
 * Change: Use inode mangler
 *
 * Revision 1.6  2000/02/16 16:57:15  jmcc
 * Autoamatically changes to the FlashPath major if a FlashPath is
 * inserted into /dev/fd0
 *
 * Revision 1.5  2000/02/14 05:47:23  jmcc
 * ioctl() style communication of the ``single /dev/fd0 concept\'\'
 *
 */

/* This is after all a module that we are making.
 */
#define MODULE	1

#include "sdlnx.h"

#define MAJOR_NR flashpath_major

#define DEVICE_NR(device) MINOR(device)
#define DEVICE_NO_RANDOM
#define DEVICE_OFF(d)
#define DEVICE_REQUEST fp_request

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/blk.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/smp_lock.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/fd.h>
#include <linux/unistd.h>

#ifdef HAVE_ASM_UACCESS_H
#include <asm/uaccess.h>
#else
#ifdef HAVE_LINUX_ISDNIF_H	/* Corel Linux fix */
#include <linux/isdnif.h>	
#else
#error Can not find the include with the copy_to_user macro
#endif
#endif

#include "sdgen.h"

EXPORT_NO_SYMBOLS;
int flashpath_major=127;
/* MODULE_PARAM(flashpath_major,"i"); */

static struct file_operations fp_fops;
static int fp_blocks[256]={0, };		/* Blocks per drive  */
static int fp_blocksizes[256]={ 0, };	/* Bytes in a block per drive */

#define DRIVE(inode)	MINOR(inode->i_rdev)
static unsigned long drive_lock=0;		/* in-use drive bitmask */
static unsigned long drive_enable=0;		/* can-use drive bitmask */

static void fixup_blocksizes(int drive)
{
	struct flashpath_drive_geom_s geom;

	flashpath_drive_get_geom(drive,&geom);
	fp_blocks[drive]=geom.sectors*geom.cylinders*geom.heads-geom.skip_sectors;
	fp_blocks[drive] >>= (BLOCK_SIZE_BITS - 9);
	fp_blocksizes[drive]=geom.sector_size;
}

/*****************************
 * Inode Manglement
 *    Here's where the fun begins. We have to modify the inode->i_rdev
 * for the inode (locking it into memory with iget()). Also, when fp_release()
 * is called, we need to _schedule_ the release of the inode 2 seconds after
 * the close is completed. 
 *
 *   This way mount(8) will work - mount will first attempt to open(2) 
 * the inode, fd_open() will mangle the inode->i_rdev, then mount(8) 
 * will close(2) the inode _immediately_ before calling sys_mount().
 * So we have a race condition - we need keep the inode in-cache until
 * the mount, BUT we to schedule the un-manglement of the inode if
 * sys_mount() is not called.
 *
 *   Therefore, on fp_release() we schedule the inode to be unmangled in 2*HZ
 * jiffies, and on fp_check_media_change() we _immediately_ unmangle the
 * inode.
 *
 * 	  Since only one instance of the flashpath device can be open at a 
 * time, we only ever need to have one tq_struct/wait_queue per drive.
 */
static struct {
	kdev_t old_dev,old_rdev;
	struct inode *inode;
	struct timer_list inode_timer;
} fp_drive_event[MAX_DRIVE];

static void fp_inode_release(int drive,long timeout);

static int fp_inode_capture(int drive,struct inode *inode)
{
	SDERROR err;
	struct inode *oinode;
	MOD_INC_USE_COUNT;

	oinode=fp_drive_event[drive].inode;
	if (oinode != NULL) {
		del_timer(&fp_drive_event[drive].inode_timer);
		clear_bit(drive,&drive_enable);
		DPRINT("fp_inode_capture(%d,%p) - already captured inode %p"
				" - releasing\n",drive,inode,oinode);
		oinode->i_dev=fp_drive_event[drive].old_dev;
		oinode->i_rdev=fp_drive_event[drive].old_rdev;
		iput(oinode);
		fp_drive_event[drive].inode=NULL;
		MOD_DEC_USE_COUNT;	/* From old acquire */
	} else {
		/* Prepare to be a flashpath..
		*/
		err=flashpath_drive_init(drive);
		if (err != SDE_SUCCESS) {
			MOD_DEC_USE_COUNT;
			return -EBUSY;
		}
	
		/* But test to see if it's a FlashPath
		 */
		err=flashpath_drive_setup(drive);
		if (err != SDE_SUCCESS) {
			flashpath_drive_cleanup(drive);
			MOD_DEC_USE_COUNT;
			return -ENXIO;
		}
	}

	inode=igrab(inode);
	if (inode == NULL) {
		printk(KERN_ERR DEVICE_NAME ":fp_inode_capture(%d,NULL) - "
					"Inode disappeared out from under me!\n",drive);
		flashpath_drive_cleanup(drive);
		MOD_DEC_USE_COUNT;
		return -ENODEV;
	}
	fp_drive_event[drive].old_dev=inode->i_dev;
	fp_drive_event[drive].old_rdev=inode->i_rdev;
	fp_drive_event[drive].inode=inode;
	inode->i_dev=MKDEV(MAJOR_NR,drive);
	inode->i_rdev=MKDEV(MAJOR_NR,drive);

	DPRINT("fp_inode_capture(%d) - captured %p (dev %.2X:%.2X)\n",
			drive,inode,
			MAJOR(inode->i_rdev),
			MINOR(inode->i_rdev));
	printk(KERN_INFO DEVICE_NAME ": Using FlashPath on /dev/fd%d\n",drive);
	return 0;
}

static void _fp_inode_release(unsigned long drive)
{
	struct inode *inode;
	
	inode=fp_drive_event[drive].inode;
	if (inode == NULL) {
		DPRINT("fp_inode_release(%ld) - already released!\n",drive);
		return;
	}

	inode->i_dev=fp_drive_event[drive].old_dev;
	inode->i_rdev=fp_drive_event[drive].old_rdev;
	iput(inode);
	fp_drive_event[drive].inode=NULL;
	clear_bit(drive,&drive_enable);
	flashpath_drive_cleanup(drive);
	DPRINT("fp_inode_release(%d) - released %p\n",(int)drive,inode);
	printk(KERN_INFO DEVICE_NAME ": Released FlashPath on /dev/fd%ld\n",drive);
	MOD_DEC_USE_COUNT;
}

/* Timeout is in jiffies
 */
static void fp_inode_release(int drive,long timeout)
{
	struct timer_list *tl;

	del_timer(&fp_drive_event[drive].inode_timer);

	/* Go ahead and do it now.
	 */
	if (timeout == 0) {		
		_fp_inode_release(drive);
		return;
	}

	/* Ok. gotta schedule it...
	 */
	tl=&fp_drive_event[drive].inode_timer;
	tl->data=drive;
	tl->function=_fp_inode_release;
	tl->expires=jiffies+timeout;

	add_timer(tl);
}

/******************
 * Event scheduler
 ******************/
static void fp_event_init(void)
{
	int i;

	for (i=0;i<MAX_DRIVE;i++) {
		fp_drive_event[i].inode=NULL;
		init_timer(&fp_drive_event[i].inode_timer);
	}
}

static void fp_event_cleanup(void)
{
	int i;

	for (i=0;i<MAX_DRIVE;i++) {
		fp_drive_event[i].inode=NULL;
		del_timer(&fp_drive_event[i].inode_timer);
	}
}

/**************************************
** Old FD operations, and switch-o-matic
***************************************/
static int (*old_fd_open)(struct inode *inode, struct file *filp);
static int (*old_fd_ioctl)(struct inode *inode, struct file *filp,
			unsigned int cmd, unsigned long arg);
static int fp_open(struct inode *inode, struct file *filp);

static int fd_open(struct inode *inode, struct file *filp)
{
	int err,drive;

	MOD_INC_USE_COUNT;

#ifdef DBG
	DPRINT("open inode %p, file %p\n",inode,filp);
	if (filp==NULL) {
		printk(KERN_ERR DEVICE_NAME ":	Argh: filp==NULL\n");
		MOD_DEC_USE_COUNT;
		return -EBUSY;
	}
	if (inode==NULL) {
		DPRINT("Argh: inode==NULL\n");
		MOD_DEC_USE_COUNT;
		return -EBUSY;
	}
#endif
	if (filp->f_flags & O_NOCTTY) {
		DPRINT("  direct\n");
		filp->f_flags &= ~O_NOCTTY;
		err=old_fd_open(inode,filp);
		MOD_DEC_USE_COUNT;
		return err;
	}

	drive=MINOR(inode->i_rdev);
	if (drive < MAX_DRIVE) {
		DPRINT("  flashpath\n");
		/* Is it a flashpath? 
		 */
		err=fp_open(inode,filp);
		if (err == 0) {
			/* Yep - better give it the right f_ops!
			 */
			filp->f_op=&fp_fops;
			MOD_DEC_USE_COUNT;
			return err;
		}
	}
	/* Nope. It's a floppy
	 */

	DPRINT("  floppy\n");
	err=old_fd_open(inode,filp);
	DPRINT("old_fd_open(%d:%d) -> %d\n",
					MAJOR(inode->i_rdev),MINOR(inode->i_rdev),err);

	MOD_DEC_USE_COUNT;
	return err;
}

static int fd_ioctl(struct inode *inode, struct file *filp,
			unsigned int cmd, unsigned long arg)
{
	int drive;

	drive=DRIVE(inode);

	if (cmd == FPSTATUS && drive < MAX_DRIVE) 
		return (int)flashpath_drive_status(drive);

	return old_fd_ioctl(inode,filp,cmd,arg);
}
		
		

/**************************************
**  FlashPath file operations
***************************************/
static int fp_open(struct inode *inode, struct file *filp) 
{
	int err;
	int drive;

	MOD_INC_USE_COUNT;

	drive=DRIVE(inode);
		
	if (filp->f_flags == (O_NDELAY|O_RDONLY)) {
		/* We only want to do IOCTLs on it
		 */
		return 0;
	}

	/* Is this drive already open?
	 */
	if (test_and_set_bit(drive,&drive_lock)) {
		MOD_DEC_USE_COUNT;
		return -EBUSY;
	}

	/* See if it's a flashpath. 
	 * We're going to have to ``capture''  the inode so that
	 * we can mangle the inode->i_rdev and other fun stuff.
	 */
	err=fp_inode_capture(drive,inode);
	if (err) {
		clear_bit(drive,&drive_lock);
		MOD_DEC_USE_COUNT;
		return err;
	}

	invalidate_buffers(inode->i_rdev);
	fixup_blocksizes(drive);

	set_device_ro(inode->i_rdev,flashpath_drive_readonly(drive));
	set_bit(drive,&drive_enable);

	return 0;
}

static int fp_release (struct inode *inode, struct file *filp) 
{
	int drive;

	drive=DRIVE(inode);
	DPRINT("release(%.2X:%.2X)\n",MAJOR(inode->i_rdev),MINOR(inode->i_rdev));

	if (filp!=NULL && filp->f_flags == (O_NDELAY|O_RDONLY)) {
		/* We only want to do IOCTLs on it
		 */
		MOD_DEC_USE_COUNT;	
		return 0;
	}

	/* Flush the write buffers - otherwise we'll get
	 * panic()s when the ll_rw_block() buffers try
	 * to flush
	 */
	fsync_dev(inode->i_rdev);
	invalidate_buffers(inode->i_rdev);

	/* Flush the FlashPath buffers
	 */
	flashpath_drive_flush(drive);

	clear_bit(drive,&drive_enable);

	/* We're done mangling this inode. We'll allow it
	 * to return to its prior state in just a few seconds..
	 */
	fp_inode_release(drive,FP_INODE_MANGLE_TIMEOUT);
	clear_bit(drive,&drive_lock);
	MOD_DEC_USE_COUNT;	
	return 0;
}

static int fp_ioctl(struct inode *inode, struct file *filp,
					unsigned int cmd, unsigned long arg)
{
	int err;
	ulong sectors;
	struct flashpath_drive_geom_s geo_flash;
	struct hd_geometry geo_fun;
	struct floppy_struct geo_flop;
	int drive;

	drive=DRIVE(inode);

	/* We know we've got a valid FlashPath drive now..
	 */
	switch (cmd) {
		case FPSTATUS:
			return (int)flashpath_drive_status(drive);
		case FDGETPRM:
			if (!arg)
				return -EINVAL;
			err = flashpath_drive_get_geom(drive,&geo_flash);
			if (err) 
				return -ENXIO;
			geo_flop.size=geo_flash.sectors*geo_flash.cylinders*geo_flash.heads-geo_flash.skip_sectors;
			geo_flop.sect=geo_flash.sectors;
			geo_flop.head=geo_flash.heads;
			geo_flop.track=geo_flash.cylinders;
			geo_flop.stretch=0;
			geo_flop.gap=27;
			geo_flop.rate=0;
			geo_flop.spec1=207;
			geo_flop.fmt_gap=108;
			return copy_to_user((long *)arg,&geo_flop,sizeof(geo_flop)) ? -EFAULT : 0;
		case BLKGETSIZE:
			if (!arg) 
				return -EINVAL;
			err = flashpath_drive_get_geom(drive,&geo_flash);
			if (err) 
				return -ENXIO;
			sectors=geo_flash.sectors*geo_flash.cylinders*geo_flash.heads-geo_flash.skip_sectors;
			return put_user(sectors,(long *)arg);
		case HDIO_GETGEO:
			err = flashpath_drive_get_geom(drive,&geo_flash);
			if (err) 
				return -ENXIO;
			geo_fun.heads 		= geo_flash.heads;
			geo_fun.sectors 	= geo_flash.sectors;
			geo_fun.cylinders	= geo_flash.cylinders;
			geo_fun.start		= 0;
			return copy_to_user((long *)arg,&geo_fun,sizeof(geo_fun)) ? -EFAULT : 0;
		default:
			break;
	}
	return -EINVAL;
}

static int fp_check_media_change(kdev_t dev)
{
	SDERROR err;

	/* Is the flashpath invalid?
	 */
	err=flashpath_drive_changed(MINOR(dev));
	if (err != SDE_SUCCESS)
		return 1;

	return 0;
}

static int fp_revalidate(kdev_t dev)
{
	SDERROR err;

	/* Revalidate (reset the drive)
	 */
	err=flashpath_drive_revalidate(MINOR(dev));
	if (err != SDE_SUCCESS)
		return -ENXIO;

	set_device_ro(dev,flashpath_drive_readonly(MINOR(dev)));

	return 0;
}

struct file_operations *fd_fops_ptr=NULL,fd_fops_old;

static struct file_operations fp_fops = {
	read:	block_read,
	write:	block_write,
	open:	fp_open,
	ioctl:	fp_ioctl,
	release:fp_release,
	check_media_change: fp_check_media_change,
	revalidate: fp_revalidate
};

/*********************
 * flashpath_request *
 *********************/
void (*fd_request)(void);

static void fp_request (void) 
{
	unsigned long processed;
	int err;
	int code;
#ifdef FP_RETRY_CODE
	int retried;
#endif
	int drive;

	while (CURRENT) {
		INIT_REQUEST;	/* do some checking on the request structure */

		drive=MINOR(CURRENT->rq_dev);

		/* For some reason, linux doesn't always
		 * flush the buffers at it should...
		 * (ie, this code can be called after a close)
		 */
		if (! test_bit(drive,&drive_enable)) {
			DPRINT("Aarrgh! Buffer request after drive closed!");
			end_request(0);
		}

		code=0;

#ifdef FP_RETRY_CODE
		retried=FP_SLEEP_TRIES;

retry:
		if (retried-- < 0) {
			end_request(0);
			continue;
		}
#endif

		switch (CURRENT->cmd) {
			case READ:
				DPRINT("READ request %ld-%ld\n",
						CURRENT->sector,
						CURRENT->sector+
						CURRENT->current_nr_sectors);
				err=flashpath_drive_read(drive,
					 CURRENT->sector,
					 CURRENT->current_nr_sectors,
					 CURRENT->buffer,&processed);
				if (err != SDE_SUCCESS) {
#ifdef FP_RETRY_CODE
					if (fp_revalidate(CURRENT->rq_dev)==0) {
						printk(KERN_NOTICE DEVICE_NAME ": device detected\n");
						goto retry;
					}
#endif
					printk(KERN_ERR DEVICE_NAME ": SLEEP ON READ! [0x%x]\n",err);
				} else 
					code=1;
				break;
			case WRITE:
				DPRINT("WRITE request %ld-%ld\n",
						CURRENT->sector,
						CURRENT->sector+
						CURRENT->current_nr_sectors);		
				err=flashpath_drive_write(drive,
					  CURRENT->sector,
					  CURRENT->current_nr_sectors,
					  CURRENT->buffer,&processed);
				if (err != SDE_SUCCESS) {
#ifdef FP_RETRY_CODE
					if (fp_revalidate(CURRENT->rq_dev)==0) {
						printk(KERN_NOTICE DEVICE_NAME ": device detected\n");
						goto retry;
					}
#endif
					printk(KERN_ERR DEVICE_NAME ": SLEEP ON WRITE! [0x%x]\n",err);
				} else {
					code=1;
				}
				break;
			default:
				printk(KERN_ERR DEVICE_NAME ": Unknown command 0x%x!",CURRENT->cmd);
				break;
		}
		end_request(code);
	}

	return;
}

/************************
 * sys_mount() override.
 ************************/
extern void *sys_call_table[];

static asmlinkage int (*old_sys_mount)(char * dev_name, char * dir_name, 
	char * type, unsigned long new_flags, void * data);

static asmlinkage int my_sys_mount(char * dev_name, char * dir_name, 
	char * type, unsigned long new_flags, void * data)
{
	int err;
	struct file *filp;
	struct dentry *dentry;

	MOD_INC_USE_COUNT;
	lock_kernel();

	DPRINT("sys_mount(%s,...)\n",dev_name);
	/* Whis is this you say? Well, if you tell mount(1)
	 * what filesystem type to look for (ie -t vfat) then
	 * mount doesn't open dev_name, which means that 
	 * sys_mount gets the wrong inode to work with.
	 *
	 * So here we take care of that by pre-opening the
	 * device if it's a floppy device.
	 */
	do {
		dentry=namei(dev_name);
		if (IS_ERR(dentry))
			break;
		DPRINT("  dentry(%p)\n",dentry);
		if (!S_ISBLK(dentry->d_inode->i_mode)) {
			dput(dentry);
			break;
		}
		DPRINT("  dev(%.2X:%.2X)\n",MAJOR(dentry->d_inode->i_rdev),
					MINOR(dentry->d_inode->i_rdev));
		if (MAJOR(dentry->d_inode->i_rdev)!=FLOPPY_MAJOR) {
			dput(dentry);
			break;
		}
		dput(dentry);
		DPRINT("  filp_open(%s)\n",dev_name);
		filp=filp_open(dev_name,O_RDONLY,0);
		if (IS_ERR(filp))
			break;
		DPRINT("  filp_close(%s)\n",dev_name);
		filp_close(filp,NULL);
	} while (0);

	unlock_kernel();
	err=old_sys_mount(dev_name,dir_name,type,new_flags,data);

	MOD_DEC_USE_COUNT;
	return err;
}

/***************
 * INIT_MODULE *
 ***************/
int init_module(void) 
{
	int result;
	struct file *filp;

	DPRINT("initializing...\n");
	result = flashpath_init();	
	if (result < 0) {
		printk(KERN_ERR DEVICE_NAME ": can't initialize\n");
		return result;
	}

	DPRINT("opening /dev/fd0\n");
	/* Take over file operations
	 */
	filp=filp_open("/dev/fd0",O_ACCMODE|O_NDELAY,0);
	if (IS_ERR(filp)) {
		printk(KERN_ERR DEVICE_NAME ": Can't get /dev/fd0 [%ld]\n",PTR_ERR(filp));
		flashpath_cleanup();
		return -ENODEV;
	}

	DPRINT("taking over /dev/fd0 (%p)\n",filp);
	fd_fops_ptr=filp->f_op;
	if (fd_fops_ptr == NULL) {
		printk(KERN_ERR DEVICE_NAME ": No floppy driver installed.\n");
		filp_close(filp,NULL);
		flashpath_cleanup();
		return -ENODEV;
	}
	DPRINT("closing fd\n");
	filp_close(filp,NULL);

	/* Pointers to block size information
	 */
	result=register_blkdev(MAJOR_NR,DEVICE_NAME,&fp_fops);
	if (result < 0) {
		printk(KERN_ERR DEVICE_NAME ": Can't register for major %d\n",
			MAJOR_NR);
		flashpath_cleanup();
		return -ENODEV;
	}

	DPRINT("overriding floppy open\n");
	memcpy(&fd_fops_old,fd_fops_ptr,sizeof(fd_fops_old));
	old_fd_open=fd_fops_ptr->open;
	old_fd_ioctl=fd_fops_ptr->ioctl;
	fd_fops_ptr->open=fd_open;
	fd_fops_ptr->ioctl=fd_ioctl;

	DPRINT("registering " DEVICE_NAME " operations\n");

	/* Take block requests
	 */
	blk_size[MAJOR_NR]=fp_blocks;
	blksize_size[MAJOR_NR]=fp_blocksizes;
	blk_dev[MAJOR_NR].request_fn=fp_request;

	DPRINT("Setting up inode mangler\n");
	fp_event_init();

	DPRINT("Modifying mount syscall [was %p, will be %p\n",
		sys_call_table[__NR_mount],my_sys_mount);
	old_sys_mount=sys_call_table[__NR_mount];
	sys_call_table[__NR_mount]=my_sys_mount;

	printk(KERN_INFO "FlashPath: Copyright 2000, SmartDisk Corp.\n");
	printk(KERN_INFO "FlashPath: Driver (v"VERSION") loaded\n");
	return 0;
}

void cleanup_module (void) 
{
	DPRINT("Modifying mount syscall [was %p, will be %p\n",
		sys_call_table[__NR_mount],old_sys_mount);
	sys_call_table[__NR_mount]=old_sys_mount;
	fp_event_cleanup();

	blk_dev[MAJOR_NR].request_fn=NULL;
	blksize_size[MAJOR_NR]=NULL;
	blk_size[MAJOR_NR]=NULL;

	unregister_blkdev(MAJOR_NR,DEVICE_NAME);

	fd_fops_ptr->open=old_fd_open;
	fd_fops_ptr->ioctl=old_fd_ioctl;

	flashpath_cleanup();

	printk(KERN_INFO "FlashPath: Driver (v"VERSION") unloaded\n");
#ifdef DBG_MEMORY
	cppfake_status();
#endif
}
