/*
 * Jason McMullan
 * Linux version of sdgen???.cpp for Flashpath 
 * $Revision: 1.17 $
 *
 * $Log: sdgenlnx.c,v $
 * Revision 1.17  2000/04/14 13:49:17  gus
 * Corel fixes
 *
 * Revision 1.16  2000/03/10 19:08:35  gus
 * Changed status reporting system - works better now
 *
 * Revision 1.15  2000/03/09 18:14:12  gus
 * Works... But I'm about to try to optimize for frequent re-opens
 *
 * Revision 1.14  2000/03/06 22:08:40  gus
 * Fixed the normal-user-cant-use-mtools bug
 *
 * Revision 1.13  2000/02/27 05:34:31  gus
 * Ahh.. Final fixes...
 *
 * Revision 1.12  2000/02/25 22:41:45  gus
 * Forgout about the duplicate inode issue...
 *
 * Revision 1.11  2000/02/25 20:29:01  gus
 * Oops..
 *
 * Revision 1.10  2000/02/24 09:51:14  gus
 * ALPHA Release. This has a working fpdaemon implementation, and a
 * 	working driver implementation. Yee ha.
 *
 * Revision 1.9  2000/02/21 23:21:52  gus
 * Finally got the floppy_raw_cmd based IO working, and it works like
 * a dream! Now performing some tests to make sure I didn't screw something up.
 *
 * Revision 1.8  2000/02/18 22:55:33  jmcc
 * Latest rev - includes all 3 types of IO, but seems to kill the
 * buffer system on too many writes...
 *
 * Revision 1.7  2000/02/17 20:14:22  jmcc
 * Change: Use inode mangler
 *
 * Revision 1.6  2000/02/16 16:57:15  jmcc
 * Autoamatically changes to the FlashPath major if a FlashPath is
 * inserted into /dev/fd0
 *
 */

#include "sdlnx.h"
#define DEVICE_NAME "flashpath"

#include <linux/kernel.h>
#include <linux/kmod.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/fd.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

/***********
 * Memory
 ***********/

/* Milliseconds to jiffies
 * Convert to kilojiffies, then jiffies
 */
#define MSEC_TO_JIFFIES(msec)	(((msec)*HZ)/1000L)

#ifndef DBG
static void *_alloc(ulong uiBytesToAllocate, SDM_MEMORY_TYPE MemoryType)
{
	return kmalloc(uiBytesToAllocate, GFP_KERNEL);
}

static void _free(void *pMemoryToFree, ulong ulSize)
{
	kfree(pMemoryToFree);
	return;
}
#else /* DBG */
static struct {
	unsigned long blocks;
	unsigned long bytes;
} _alloc_info= {0,0};

static void *_alloc(ulong uiBytesToAllocate, SDM_MEMORY_TYPE MemoryType)
{
	unsigned long *cp;
	
	cp=(unsigned long *)kmalloc(uiBytesToAllocate+sizeof(unsigned long),
				   				GFP_KERNEL);
	if (cp==NULL)
		return NULL;

	_alloc_info.blocks++;
	_alloc_info.bytes+=uiBytesToAllocate;
	*(cp++)=uiBytesToAllocate;

	return (void *)cp;
}

static void _free(void *pMemoryToFree, ulong ulSize)
{
	unsigned long *cp=(unsigned long *)pMemoryToFree;

	if (cp==NULL)
		return;

	cp--;
	_alloc_info.blocks--;
	_alloc_info.bytes-=(*cp);
	kfree(cp);
	return;
}

static void _alloc_status(void)
{
		DPRINT("Allocation:\n");
		cppfake_status();
		DPRINT(" SDGEN: %ld blocks, %ld bytes\n",
					_alloc_info.blocks,_alloc_info.bytes);
}
#endif

/***********
 * Time
 ***********/
static void _get(long *sec,long *usec)
{
	struct timeval tv;

	get_fast_time(&tv);
	(*sec)=tv.tv_sec;
	(*usec)=tv.tv_usec;
}

/***********
 * Tracing
 ***********/
void _vnotify(TRACE_LEVEL level,const char *buff,va_list args)
{
  char buffer[1024];
  int i;
  static const struct {
		TRACE_LEVEL level;
		int	klevel;
  } level_map[] = {
		{TraceAlways,1},
		{TraceInfo,7},
		{TraceWarning,4},
		{TraceError,3},
		{TraceFatal,2},
		{TraceNever,99}};	/* This one must always be last */

  // Should really be vsnprintf, but Linux Kernel 2.2.x doesn't have it. 
  vsprintf(buffer,buff, args);

  for (i=0;level_map[i].level != TraceNever; i++)
	if (level_map[i].level==level)
		break;

  printk("<%d>" DEVICE_NAME ": %s",level_map[i].klevel,buffer);
}

void _notify(TRACE_LEVEL level,const char *buff,...)
{
  va_list these;
  va_start(these, buff);

  _vnotify(level,buff,these);
}

void _panic(const char *buff,...)
{
	char buffer[1024];
	va_list these;
	va_start(these, buff);

  	// Should really be vsnprintf, but Linux Kernel 2.2.x doesn't have it. 
	vsprintf(buffer,buff, these);

	panic(buffer);
}

/***********
 * Floppy
 ***********/

#ifdef DBG
static void _dump_sector(unsigned long addr,const unsigned char *b)
{
	int i,j;
	for (i=0;i<0x100;i+=16,b+=16) {
		printk(KERN_ERR "0x%.8lX: %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X  %2X %2X %2X %2X %2X %2X %2X %2X ",
		addr+i,*b, b[1], b[2], b[3], b[4], b[5], b[6], b[7],
		b[8], b[9], b[10], b[11], b[12], b[13], b[14], b[15]);
		for (j=0;j<16;j++) {
			if (b[j] >= 32 && b[j] <=126) {
				printk("%c",b[j]);
			} else {
				printk(".");
			}
		}
		printk("\n");
	}
}
#endif


static struct {
	struct file file;
	struct inode *inode;
} _fp_drive[MAX_DRIVE] = {{ {}, NULL}, };

#define GETFILE(h,retval) \
	int drive; \
	struct file *file; \
	struct inode *inode; \
	drive=((char *)h-(char *)NULL)-1; \
	if (drive < 0 || drive >= MAX_DRIVE) { \
		printk(KERN_ERR DEVICE_NAME ": ignoring bad handle %p at %s:%d\n", \
			h,__FILE__,__LINE__); \
		return retval; } \
	inode=_fp_drive[drive].inode; \
	if (inode == NULL) { \
		printk(KERN_ERR DEVICE_NAME ": inactive drive %d at %s:%d\n",\
			drive,__FILE__,__LINE__); \
		return retval; } \
	file=&_fp_drive[drive].file; \

#define BSIZE	512

static void *_open_by_id(int drive)
{
	struct file *filp;
	struct inode *inode;
	int err;

	if (_fp_drive[drive].inode != NULL) {
		printk(KERN_ERR DEVICE_NAME ": Drive %d is currently in use\n",
				drive);
		return NULL;
	}
	
	/* Create a new inode, and make its device
	 * to be FLOPPY_MAJOR, and drive.
	 */
	inode=get_empty_inode();
	inode->i_dev=inode->i_rdev=MKDEV(FLOPPY_MAJOR,drive);
	insert_inode_hash(inode);

	/* Now the file pointer...
	 */	
	filp=&_fp_drive[drive].file;
	memset(filp,0,sizeof(struct file));
	filp->f_op=fd_fops_ptr;
	filp->f_flags=O_RDWR|O_NDELAY|O_NOCTTY;
	filp->f_mode=2;

	/* And finally, do a blkdev_open to seal the deal.
	 */
	err=blkdev_open(inode,filp);
	if (err) {
		printk(KERN_ERR DEVICE_NAME ": Can't open /dev/fd%dH1440\n",
				drive);
		clear_inode(inode);
		_fp_drive[drive].inode=NULL;
	}	

	set_blocksize(inode->i_rdev,BSIZE);
	_fp_drive[drive].inode=inode;

	return (void *)(drive+1);
}

static void _close(void *h)
{
	GETFILE(h,);

	DPRINT("Closing device %.4X\n",inode->i_rdev);
	blkdev_release(inode);
	clear_inode(inode);
	_fp_drive[drive].inode=NULL;
}

static SDERROR _configure(void *h)
{ 
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	/* Nothing to be done -- already configured?? */ 
	return SDE_SUCCESS;
}

static SDERROR _deviceiocontrol (void *h, uint uiDeviceControlCode,
			       PVOID pInputBuffer,
			       uint uiInputBufferLength,
			       PVOID pOutputBuffer,
			       uint uiOutputBufferLength,
			       uint * puiActualOutputLength )
{ 
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	return SDE_SUCCESS;
	/* Don't know what to do with this yet -- nothing so far*/ 
}

#if (FP_IO_MODE == FP_IO_RAW)
/* Now we initialize the skeleton raw commands */
#define DRIVE	cmd[1]
#define TRACK	cmd[2]
#define	SECTOR	cmd[4]
#define	TRACK_LN cmd[6]

static unsigned long drive_flush=0;	/* Do we need to flush? */

/* Reading command */
struct floppy_raw_cmd read_f = {
	flags:	FD_RAW_READ|FD_RAW_INTR,
	cmd_count:	9,
  	cmd: { 0x66, 0x00, 0x00, 0x00, 0x00, 0x02, 0x12, 0x1b, 0xff }
};
/* Writing command */
struct floppy_raw_cmd write_f = {
	flags:	FD_RAW_WRITE|FD_RAW_INTR,
	cmd_count:	9,
  	cmd: { 0x45, 0x00, 0x00, 0x00, 0x00, 0x02, 0x12, 0x1b, 0xff }
};
/* Seek track command */
struct floppy_raw_cmd seek_f = {
	flags:	FD_RAW_INTR,
	cmd_count:	3,
  	cmd: { 0x0f, 0x00, 0x00 }
};
/* Drive Status command */
struct floppy_raw_cmd status_f = {
	flags: 		0,
	cmd_count:	2,
	cmd: { 0x04, 0x00 }
};
/* Recalibrate command (to be called after a reset) */
struct floppy_raw_cmd recal_f = {
	flags: FD_RAW_INTR,
	cmd_count:	2,
	cmd: { 0x07, 0x00 }
};

/***************
 * Read Buffer *
 ***************/
static SDERROR _read(void *h,
			ulong ulLba, ulong ulBytesToRead, ulong ulMaxWait, 
			uchar * pBuffer, uint track_length)
{
	int res;
	int start,len;
	unsigned long end;	/* Timer */
	struct floppy_raw_cmd command;
	mm_segment_t old_fs;
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	/* Seek... */
	if (ulLba%BSIZE || ulBytesToRead%BSIZE) {
			printk(KERN_ERR DEVICE_NAME ": Can't use wacky offset,len of 0x%lx,0x%lx!\n",ulLba,ulBytesToRead);
			return SDE_SECTOR_NOT_FOUND;
	}

	start=(ulLba/BSIZE)+1;
	len=ulBytesToRead/BSIZE;

	/* First, invalidate the floppy driver's track buffer. */
	if (test_and_clear_bit(drive,&drive_flush))
		file->f_op->ioctl(inode,file,FDFLUSH,0);

	/* Start timer */
	end=jiffies+MSEC_TO_JIFFIES(ulMaxWait);

	DPRINT("Read %ld bytes at %ld\n",ulBytesToRead,ulLba);

	/* Now read the blocks */
	while (! time_after(jiffies,end)) {
		/* Read from the floppy. */
		command = read_f;
		command.data=pBuffer;
		command.length=ulBytesToRead;
		command.SECTOR=start;
		command.DRIVE=drive & 0x3;	/* Only 2 bits */
		if (track_length != 0)
			command.TRACK_LN=track_length;

		/* Magic to be able to use ioctl() on kernel memory */
		old_fs=get_fs();
		set_fs(get_ds());

		res=file->f_op->ioctl(inode,file,FDRAWCMD,(unsigned long)&command);

		set_fs(old_fs);

		DPRINT("Read status: %d\n",res);

		if (command.reply[0] < 7) {
			/* Read was successful */
			return SDE_SUCCESS;
		}
		/* Dang. try again... */
	}

	printk(KERN_ERR DEVICE_NAME ": Read failed at %ld\n",ulLba);
	return SDE_SECTOR_NOT_FOUND;
}

/****************
 * Write Buffer
 *****************/
static SDERROR _write(void *h,
			ulong ulLba, ulong ulBytesToWrite, ulong ulMaxWait, 
			uchar * pBuffer)
{
	int res;
	int start,len;
	unsigned long end;
	struct floppy_raw_cmd command;
	mm_segment_t old_fs;
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	/* Seek... */
	if (ulLba%BSIZE || ulBytesToWrite%BSIZE) {
			printk(KERN_ERR DEVICE_NAME ": Can't use wacky offset,len of 0x%lx,0x%lx!\n",ulLba,ulBytesToWrite);
			return SDE_SECTOR_NOT_FOUND;
	}

	start=(ulLba/BSIZE)+1;
	len=ulBytesToWrite/BSIZE;

	/* Start timer */
	end=jiffies+MSEC_TO_JIFFIES(ulMaxWait);

	set_bit(drive,&drive_flush);
	DPRINT("Write %ld bytes at %ld\n",ulBytesToWrite,ulLba);

	/* Now write the blocks */
	while (! time_after(jiffies,end)) {
		/* Read from the floppy. */
		command = write_f;
		command.data=pBuffer;
		command.length=ulBytesToWrite;
		command.SECTOR=start;
		command.DRIVE=drive & 0x03; /* Only 2 bits */

		/* Magic to be able to use ioctl() on kernel memory */
		old_fs=get_fs();
		set_fs(get_ds());

		res=file->f_op->ioctl(inode,file,FDRAWCMD,(unsigned long)&command);

		set_fs(old_fs);

		DPRINT("Write status: %d\n",res);

		if (command.reply[0] < 7) {
			/* Write was successful */
			return SDE_SUCCESS;
		}
		/* Dang. try again... */
	}

	printk(KERN_ERR DEVICE_NAME ": Write failed at %ld\n",ulLba);
	return SDE_SECTOR_NOT_FOUND;
}
#endif /* FP_IO_RAW */
#if (FP_IO_MODE == FP_IO_READWRITE)
/***************
 * Read Buffer *
 ***************/
static SDERROR _read(void *h,
			ulong ulLba, ulong ulBytesToRead, ulong ulMaxWait, 
			uchar * pBuffer, uint track_length)
{
	int res;
	loff_t offset;
	mm_segment_t old_fs;
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	do_lseek(file,ulLba,0);
	offset=file->f_pos;

	file->f_op->ioctl(inode,file,FDFLUSH,0);

	/* Since we're in the ll_rw_block() request
	 * handler, we'll need to unlock the io_request_lock
	 * to be able to do any more reads or writes.
	 */
	spin_unlock_irq(&io_request_lock);
	old_fs=get_fs();
	set_fs(get_ds());
	res=file->f_op->read(file,pBuffer,ulBytesToRead,&offset);
	set_fs(old_fs);
	spin_lock_irq(&io_request_lock);

	return (res >= 0) ? 0 : SDE_SECTOR_NOT_FOUND;
}

/****************
 * Write Buffer *
 ****************/
static SDERROR _write(void *h,
			ulong ulLba, ulong ulBytesToWrite, 
			ulong ulMaxWait, uchar * pBuffer)
{
	int res;
	loff_t offset;
	mm_segment_t old_fs;
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	do_lseek(file,ulLba,0);
	offset=file->f_pos;

	/* Since we're in the ll_rw_block() request
	 * handler, we'll need to unlock the io_request_lock
	 * to be able to do any more reads or writes.
	 */
	spin_unlock_irq(&io_request_lock);
	old_fs=get_fs();
	set_fs(get_ds());
	res=file->f_op->write(file,pBuffer,ulBytesToWrite,&offset);
	set_fs(old_fs);
	spin_lock_irq(&io_request_lock);

	fsync_dev(node->i_rdev);

	return (res >= 0) ? 0 : SDE_SECTOR_NOT_FOUND;
}
#endif /* FP_IO_READWRITE */
#if (FP_IO_MODE == FP_IO_BLOCK)
/***************
 * Read Buffer *
 ***************/
static SDERROR _read(void *h,
			ulong ulLba, ulong ulBytesToRead, ulong ulMaxWait, 
			uchar * pBuffer, uint track_length)
{
	int i,start,len;
	kdev_t dev;
	struct buffer_head *bh=NULL;
	unsigned long end;	/* Timer */
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

#ifdef DBG
	_alloc_status();
#endif

	/* Seek...
	 */
	if (ulLba%BSIZE || ulBytesToRead%BSIZE) {
			printk(KERN_ERR DEVICE_NAME ": Can't use wacky offset,len of 0x%lx,0x%lx!\n",ulLba,ulBytesToRead);
			return SDE_SECTOR_NOT_FOUND;
	}
	start=ulLba/BSIZE;
	len=ulBytesToRead/BSIZE;
	dev=inode->i_rdev;

	DPRINT("Read %ld bytes at %ld\n",ulBytesToRead,ulLba);

	/* Start timer
	 */
	end=jiffies+MSEC_TO_JIFFIES(ulMaxWait);

	/* Since we're in the ll_rw_block() request
	 * handler, we'll need to unlock the io_request_lock
	 * to be able to do any more reads or writes.
	 */
	spin_unlock_irq(&io_request_lock);
	for (i=0;i<len;i++) {
		bh=bread(dev,start+i, BSIZE);
		if (bh==NULL)
			break;
#if 0
		/* Oops. Took too long...
		 */
		if (time_after(jiffies,end)) {
			brelse(bh);
			bh=NULL;
			break;
		}
#endif
		/*_dump_sector(i,bh->b_data);*/
		memcpy(pBuffer+i*BSIZE,bh->b_data,BSIZE);
		brelse(bh);
	}
	spin_lock_irq(&io_request_lock);

	DPRINT("Read %d bytes\n",i*BSIZE);

	return (bh != NULL) ? 0 : SDE_SECTOR_NOT_FOUND;
}

/****************
 * Write Buffer *
 ****************/
static SDERROR _write(void *h,
			ulong ulLba, ulong ulBytesToWrite, ulong ulMaxWait, 
			uchar * pBuffer)
{
	int i,start,len;
	kdev_t dev;
	struct buffer_head *bh;
	unsigned long end;
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

#ifdef DBG
	_alloc_status();
#endif

	/* Seek... 
	 */
	if (ulLba%BSIZE || ulBytesToWrite%BSIZE) {
			printk(KERN_ERR DEVICE_NAME ": Can't use wacky offset,len of 0x%lx,0x%lx!\n",ulLba,ulBytesToWrite);
			return SDE_SECTOR_NOT_FOUND;
	}
	start=ulLba/BSIZE;
	len=ulBytesToWrite/BSIZE;

	dev=inode->i_rdev;

	DPRINT("Write %ld bytes at %ld\n",ulBytesToWrite,ulLba);

	/* Start timer
	 */
	end=jiffies+MSEC_TO_JIFFIES(ulMaxWait);

	/* Since we're in the ll_rw_block() request
	 * handler, we'll need to unlock the io_request_lock
	 * to be able to do any more reads or writes.
	 */
	spin_unlock_irq(&io_request_lock);
	bh=NULL;
	for (i=0;i<len;i++) {
#if 0
		if (time_after(jiffies,end))
			break;
#endif

		/* Set up buffer head
		 */
		bh=getblk(dev,start+i, BSIZE);
		if (bh==NULL) {
			DPRINT("[%d] Can't get buffer head\n",i+start);
			continue;
		}
		memcpy(bh->b_data,pBuffer+i*BSIZE,BSIZE);
		/*_dump_sector(i,bh->b_data); */
		mark_buffer_dirty(bh,1);
		DPRINT("[%d] ll_rw_block\n",i+start);
		ll_rw_block(WRITE,1,&bh);
		DPRINT("[%d] wait_on_buffer\n",i+start);
		wait_on_buffer(bh);
		DPRINT("[%d] brelse\n",i+start);
		brelse(bh);
	}
	spin_lock_irq(&io_request_lock);

	DPRINT("Wrote %d bytes\n",i*BSIZE);

	/* Last, invalidate the floppy driver's track buffer.
	 */
	file->f_op->ioctl(inode,file,FDFLUSH,0);

	return (bh != NULL) ? 0 : SDE_SECTOR_NOT_FOUND;
}
#endif /* FP_IO_BLOCK */

/**************************
 * Check for Media Change *
 **************************/
static BOOLEAN _media_changed_supported(void *h) 
{
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	return (file->f_op->check_media_change != NULL) ? TRUE : FALSE;
}

#if (FP_IO_MODE == FP_IO_RAW)
static SDERROR _media_changed(void *h, uint * puiChangeCount )  
{
	SDERROR retval = SDE_SUCCESS;
	struct floppy_raw_cmd curr;
	mm_segment_t old_fs;
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	curr = status_f;
	curr.DRIVE=drive & 0x03; /* Only 2 bits */
	old_fs=get_fs();
	set_fs(get_ds());
	file->f_op->ioctl(inode,file,FDRAWCMD, (unsigned long)&curr);
	set_fs(old_fs);

	if (curr.flags & FD_RAW_DISK_CHANGE)
		retval = SDE_MEDIA_CHANGE_DETECTED;
	return (retval);
}

/*************
 * Reset FDC *
 *************/
static void _reset(void *h)
{
	GETFILE(h,);

	file->f_op->ioctl(inode,file,FDRESET,FD_RESET_ALWAYS);
}

/**********************
 * Recalibrate Floppy *
 **********************/
static int _revalidate(void *h) 
{
	struct floppy_raw_cmd curr, curr1;
	mm_segment_t old_fs;
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	curr = seek_f;
	curr.DRIVE=drive & 0x03; /* Only 2 bits */
	curr.TRACK = 1;
	curr.next = &curr1;
	
	curr1 = seek_f; 
	curr1.DRIVE=drive & 0x03; /* Only 2 bits */
	curr1.TRACK = 0;

	old_fs=get_fs();
	set_fs(get_ds());
	file->f_op->ioctl(inode,file,FDRESET, FD_RESET_ALWAYS);
	file->f_op->ioctl(inode,file,FDRAWCMD, (unsigned long)&curr);
	set_fs(old_fs);

	return SDE_SUCCESS;
}
#endif /* FP_IO_RAW */
#if ( FP_IO_MODE == FP_IO_READWRITE ) || (FP_IO_MODE == FP_IO_BLOCK)
static SDERROR _media_changed(void *h, uint * puiChangeCount )  
{  
	int res;
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	res=file->f_op->check_media_change(inode->i_rdev);

	return res ? SDE_MEDIA_CHANGE_DETECTED : SDE_SUCCESS;
}

/*************
 * Reset FDC *
 *************/
static void _reset(void *h)
{
	GETFILE(h,);

	if (file->f_op->ioctl) {
		file->f_op->ioctl(inode,file,FDRESET,FD_RESET_ALWAYS);
	}
}

/**********************
 * Recalibrate Floppy *
 **********************/
static int _revalidate(void *h) 
{
	GETFILE(h,SDE_NO_MEDIA_IN_DEVICE);

	/* Not sure if I can do this on an open drive...
	 */
	if (file->f_op->revalidate) {
		file->f_op->revalidate(inode->i_rdev);
	}

	return SDE_SUCCESS;
}
#endif /* FP_IO_BLOCK || FP_IO_READWRITE */

/**********************************
 * Initialize The SDGEN structure *
 **********************************/

struct SDGEN_s SDGEN={
		/* Trace */
		trace:{
			notify:_notify,
			vnotify:_vnotify,
			panic:_panic
		},
		time:{
			get:_get
		},
		mem:{
			alloc:_alloc,
			free:_free
		},
		floppy:{
			open:_open_by_id,
			close:_close,
			configure:_configure,
			deviceiocontrol:_deviceiocontrol,
			read:_read,
			write:_write,
			media_changed_supported:_media_changed_supported,
			media_changed:_media_changed,
			reset:_reset,
			revalidate:_revalidate
		}
};
